﻿using Infragistics.DragDrop;
using Infragistics.Windows;
using Infragistics.Windows.Controls;
using Infragistics.Windows.DataPresenter;
using Infragistics.Windows.DataPresenter.Events;

using MoreLinq.Extensions;

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Input;
using System.Windows.Threading;

namespace WpfApp1
{
    /// <summary>
    /// Interaction logic for MainWindow.xaml
    /// </summary>
    public partial class MainWindow
    {
        public MainWindow()
        {
            InitializeComponent();

            // set up workspaces...
            var workspaces = new List<Workspace>
                                 {
                                     new Workspace(id: 1),
                                     new Workspace(id: 2),
                                     new Workspace(id: 3),
                                     new Workspace(id: 4)
                                 };

            Workspaces = workspaces;

            // generate the raw data
            RawData = DataGenerator.GetData(count: 30);

            // select the first workspace...
            CurrentWorkspace = workspaces.First();

            CurrentWorkspace.AllowRowFiltering = false;

            // trigger a default save after being loaded
            Loaded += OnLoaded;
        }

        private void OnLoaded(object sender, RoutedEventArgs e)
        {
            var grid = this.FindChildType<XamDataGrid>();

            using var ms = new MemoryStream();

            grid.SaveCustomizations(ms, saveFieldPositionInfoAlways: true);

            DefaultLayout = Encoding.UTF8.GetString(ms.ToArray());
        }

        private string DefaultLayout { get; set; }

        #region Workspaces Readonly Dependency Property

        internal static readonly DependencyPropertyKey WorkspacesPropertyKey =
            DependencyProperty.RegisterReadOnly("Workspaces", typeof(IReadOnlyList<Workspace>), typeof(MainWindow), new PropertyMetadata(default(IReadOnlyList<Workspace>)));

        public static readonly DependencyProperty WorkspacesProperty = WorkspacesPropertyKey.DependencyProperty;

        public IReadOnlyList<Workspace> Workspaces
        {
            get
            {
                IReadOnlyList<Workspace> res = default(IReadOnlyList<Workspace>);

                if (Dispatcher?.CheckAccess() != false)
                    res = (IReadOnlyList<Workspace>)GetValue(WorkspacesProperty);
                else
                    Dispatcher?.Invoke(() => res = (IReadOnlyList<Workspace>)GetValue(WorkspacesProperty));

                return res;
            }
            private init
            {
                if (Dispatcher?.CheckAccess() != false)
                    SetValue(WorkspacesPropertyKey, value);
                else
                    Dispatcher?.Invoke(() => SetValue(WorkspacesPropertyKey, value));
            }
        }

        #endregion

        #region RawData Dependency Property

        public static readonly DependencyProperty RawDataProperty =
            DependencyProperty.Register("RawData",
                                        typeof(IEnumerable<Data>),
                                        typeof(MainWindow),
                                        new PropertyMetadata(default(IEnumerable<Data>))
                                       );

        public IEnumerable<Data> RawData
        {
            get
            {
                IEnumerable<Data> res = default(IEnumerable<Data>);

                if (CheckAccess() != false)
                    res = (IEnumerable<Data>)GetValue(RawDataProperty);
                else
                    Dispatcher?.Invoke(() => res = (IEnumerable<Data>)GetValue(RawDataProperty));

                return res;
            }
            init
            {
                if (CheckAccess() != false)
                    SetValue(RawDataProperty, value);
                else
                    Dispatcher?.Invoke(() => SetValue(RawDataProperty, value));
            }
        }

        #endregion

        #region FlatDataRows Readonly Dependency Property

        internal static readonly DependencyPropertyKey FlatDataRowsPropertyKey =
            DependencyProperty.RegisterReadOnly("FlatDataRows", typeof(IEnumerable<DataRow>), typeof(MainWindow), new PropertyMetadata(default(IEnumerable<DataRow>), OnDataChanged));

        private static void OnDataChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            if (d is MainWindow window)
                window.Data = null;
        }

        public static readonly DependencyProperty FlatDataRowsProperty = FlatDataRowsPropertyKey.DependencyProperty;

        public IEnumerable<DataRow> FlatDataRows
        {
            get
            {
                IEnumerable<DataRow> res = default(IEnumerable<DataRow>);

                if (Dispatcher?.CheckAccess() != false)
                    res = (IEnumerable<DataRow>)GetValue(FlatDataRowsProperty);
                else
                    Dispatcher?.Invoke(() => res = (IEnumerable<DataRow>)GetValue(FlatDataRowsProperty));

                return res;
            }
            private set
            {
                if (Dispatcher?.CheckAccess() != false)
                    SetValue(FlatDataRowsPropertyKey, value);
                else
                    Dispatcher?.Invoke(() => SetValue(FlatDataRowsPropertyKey, value));
            }
        }

        #endregion

        #region CompositeData Readonly Dependency Property

        internal static readonly DependencyPropertyKey CompositeDataPropertyKey =
            DependencyProperty.RegisterReadOnly("CompositeData", typeof(IEnumerable<DataRow>), typeof(MainWindow), new PropertyMetadata(default(IEnumerable<DataRow>), OnDataChanged));

        public static readonly DependencyProperty CompositeDataProperty = CompositeDataPropertyKey.DependencyProperty;

        public IEnumerable<DataRow> CompositeData
        {
            get
            {
                IEnumerable<DataRow> res = default(IEnumerable<DataRow>);

                if (Dispatcher?.CheckAccess() != false)
                    res = (IEnumerable<DataRow>)GetValue(CompositeDataProperty);
                else
                    Dispatcher?.Invoke(() => res = (IEnumerable<DataRow>)GetValue(CompositeDataProperty));

                return res;
            }
            private set
            {
                if (Dispatcher?.CheckAccess() != false)
                    SetValue(CompositeDataPropertyKey, value);
                else
                    Dispatcher?.Invoke(() => SetValue(CompositeDataPropertyKey, value));
            }
        }

        #endregion

        #region Data Readonly Dependency Property

        internal static readonly DependencyPropertyKey DataPropertyKey =
            DependencyProperty.RegisterReadOnly("Data",
                                                typeof(IEnumerable<DataRow>),
                                                typeof(MainWindow),
                                                new PropertyMetadata(default(IEnumerable<DataRow>),
                                                                     propertyChangedCallback: null,
                                                                     CoerceData
                                                                    )
                                               );

        private static object CoerceData(DependencyObject d, object baseValue)
        {
            IEnumerable<DataRow> data = Array.Empty<DataRow>();

            if (d is MainWindow window)
                data = window.SelectDataSource() ?? data;

            return data;
        }

        private IEnumerable<DataRow> SelectDataSource()
        {
            var grid = this.FindChildType<XamDataGrid>();

            return (grid?.GroupByAreaMulti?.FieldLayoutGroupByInfos.Count ?? 0) > 0 ? FlatDataRows : CompositeData;
        }

        public static readonly DependencyProperty DataProperty = DataPropertyKey.DependencyProperty;

        public IEnumerable<DataRow> Data
        {
            get
            {
                IEnumerable<DataRow> res = default(IEnumerable<DataRow>);

                if (Dispatcher?.CheckAccess() != false)
                    res = (IEnumerable<DataRow>)GetValue(DataProperty);
                else
                    Dispatcher?.Invoke(() => res = (IEnumerable<DataRow>)GetValue(DataProperty));

                return res;
            }
            private set
            {
                if (Dispatcher?.CheckAccess() != false)
                    SetValue(DataPropertyKey, value);
                else
                    Dispatcher?.Invoke(() => SetValue(DataPropertyKey, value));
            }
        }

        #endregion

        #region CurrentWorkspace Dependency Property

        public static readonly DependencyProperty CurrentWorkspaceProperty =
            DependencyProperty.Register("CurrentWorkspace",
                                        typeof(Workspace),
                                        typeof(MainWindow),
                                        new PropertyMetadata(default(Workspace), CurrentWorkspaceChanged)
                                       );

        private void SetData(Workspace workspace)
        {
            if (workspace != null)
            {
                // pick out the active workspace items
                var dataRows = RawData.Where(x => x.Workspaces.Contains(workspace.WorkspaceId)).ToArray();

                Dispatcher.InvokeAsync(() =>
                                              {
                                                  var dataGrid = this.FindChildType<XamDataGrid>();
                                                  dataGrid?.LoadCustomizations(string.IsNullOrWhiteSpace(workspace.LayoutCache) == false ? workspace.LayoutCache : DefaultLayout);
                                                  Data = null;
                                              }
                                             );

                // clear out any stored data
                FlatDataRows = CompositeData = Array.Empty<DataRow>();

                int i = 0;

                // set up the flat rows from the orignal data set
                FlatDataRows = dataRows.Select(x => new DataRow
                                                               {
                                                                   Item = x, RawIndex = i++
                                                               }
                                                     )
                                              .ToArray();

                i = 0;
                var items = dataRows.Select(x => new DataRow
                                                     {
                                                         Item = x, RawIndex = i++
                                                     }
                                           )
                                    .ToList();

                var composites = items.Where(x => x.Item.IsComposite).ToArray();

                foreach (var c in composites)
                {
                    var speeds = items.Where(x => (x.Item.IsComposite == false) &&
                                                  (c.Item.File == x.Item.File) &&
                                                  c.Item.Workspaces.Contains(workspace.WorkspaceId)
                                            )
                                      .OrderBy(x => x.RawIndex)
                                      .ToArray();
                    // remove runs from the top level
                    speeds.ForEach(x => items.Remove(x));

                    // and set children of the composite items...
                    c.Children = speeds;
                }

                // set up the composite version of the data...
                CompositeData = items;
            }
        }

        private static void CurrentWorkspaceChanged(DependencyObject obj, DependencyPropertyChangedEventArgs e)
        {
            if (obj is MainWindow window)
            {
                if (e.NewValue is Workspace workspace)
                    window.SetData(workspace);
            }
        }

        public Workspace CurrentWorkspace
        {
            get
            {
                Workspace res = default(Workspace);

                if (CheckAccess() != false)
                    res = (Workspace)GetValue(CurrentWorkspaceProperty);
                else
                    Dispatcher?.Invoke(() => res = (Workspace)GetValue(CurrentWorkspaceProperty));

                return res;
            }
            set
            {
                if (CheckAccess() != false)
                    SetValue(CurrentWorkspaceProperty, value);
                else
                    Dispatcher?.Invoke(() => SetValue(CurrentWorkspaceProperty, value));
            }
        }

        #endregion

        #region InDragMode Readonly Dependency Property

        internal static readonly DependencyPropertyKey InDragModePropertyKey =
            DependencyProperty.RegisterReadOnly("InDragMode", typeof(bool), typeof(MainWindow), new PropertyMetadata(default(bool), OnDragModeChanged));

        private static void OnDragModeChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            if (d is MainWindow window)
            {
                var grid = window.FindChildType<XamDataGrid>();

                if ((grid != null) && (e.NewValue is bool value))
                {
                    // set properties based on mode
                    if(value)
                    {
                        grid.FieldLayoutSettings.SelectionTypeRecord = SelectionType.None;
                        grid.FieldSettings.CellClickAction = CellClickAction.Default;
                    }
                    else
                    {
                        grid.FieldLayoutSettings.SelectionTypeRecord = SelectionType.ExtendedAutoDrag;
                        grid.FieldSettings.CellClickAction = CellClickAction.SelectRecord;
                    }
                }
            }
        }

        public static readonly DependencyProperty InDragModeProperty = InDragModePropertyKey.DependencyProperty;

        public bool InDragMode
        {
            get
            {
                bool res = default(bool);

                if (Dispatcher?.CheckAccess() != false)
                {
                    res = (bool)GetValue(InDragModeProperty);
                }
                else
                {
                    Dispatcher?.Invoke(() => res = (bool)GetValue(InDragModeProperty));
                }

                return res;
            }
            private set
            {
                if (Dispatcher?.CheckAccess() != false)
                {
                    SetValue(InDragModePropertyKey, value);
                }
                else
                {
                    Dispatcher?.Invoke(() => SetValue(InDragModePropertyKey, value));
                }
            }
        }

        #endregion

        #region ToggleFilter command

        private ICommand mToggleFilterCommand;

        public ICommand ToggleFilterCommand => mToggleFilterCommand ??= new RelayCommand(ToggleFilterExecute, CanToggleFilterExecute);

        private void ToggleFilterExecute(object arg)
        {
            if (CurrentWorkspace != null)
            {
                var grid = this.FindChildType<XamDataGrid>();

                if (grid?.FieldSettings != null)
                    grid.FieldSettings.AllowRecordFiltering = CurrentWorkspace.AllowRowFiltering = CurrentWorkspace.AllowRowFiltering == false;
            }
        }

        private bool CanToggleFilterExecute(object arg)
        {
            var grid = this.FindChildType<XamDataGrid>();

            return grid != null &&  CurrentWorkspace != null;
        }

        #endregion

        #region RemoveItem command

        private ICommand mRemoveItemCommand;

        public ICommand RemoveItemCommand => mRemoveItemCommand ??= new RelayCommand(RemoveItemExecute, CanRemoveItemExecute);

        private void RemoveItemExecute(object arg)
        {
        }

        private bool CanRemoveItemExecute(object arg)
        {
            var grid = this.FindChildType<XamDataGrid>();
            return (grid?.SelectedDataItems?.Length ?? 0) > 0;
        }

        #endregion

        #region Overrides of UIElement

        /// <summary>Invoked when an unhandled <see cref="E:System.Windows.Input.Keyboard.KeyDown" /> attached event reaches an element in its route that is derived from this class. Implement this method to add class handling for this event.</summary>
        /// <param name="e">The <see cref="T:System.Windows.Input.KeyEventArgs" /> that contains the event data.</param>
        protected override void OnPreviewKeyDown(KeyEventArgs e)
        {
            if ((e.Key == Key.System) || (e.Key == Key.LeftAlt) || (e.Key == Key.RightAlt))
                InDragMode = true;

            base.OnPreviewKeyDown(e);
        }

        /// <summary>Invoked when an unhandled <see cref="E:System.Windows.Input.Keyboard.KeyUp" /> attached event reaches an element in its route that is derived from this class. Implement this method to add class handling for this event.</summary>
        /// <param name="e">The <see cref="T:System.Windows.Input.KeyEventArgs" /> that contains the event data.</param>
        protected override void OnPreviewKeyUp(KeyEventArgs e)
        {
            if ((e.Key == Key.System) || (e.Key == Key.LeftAlt) || (e.Key == Key.RightAlt))
                InDragMode = false;

            base.OnPreviewKeyUp(e);
        }

        #endregion

        private void OnGrouping(object sender, GroupingEventArgs e)
        {
            Dispatcher.InvokeAsync(() =>
                                   {
                                       SaveCurrentWorkspaceLayout();

                                       Data = null;
                                   }
                                  );
        }

        private void DataGrid_OnAssigningFieldLayoutToItem(object sender, AssigningFieldLayoutToItemEventArgs e)
        {
            e.FieldLayout = e.ParentExpandableFieldRecord == null ? (sender as XamDataGrid)?.FieldLayouts["ParentLayout"] : (sender as XamDataGrid)?.FieldLayouts["ChildLayout"];
        }

        private void DataGrid_OnFieldPositionChanged(object sender, FieldPositionChangedEventArgs e)
        {
            if (sender is XamDataGrid dataGrid)
            {
                if (dataGrid.FieldLayouts[index: 0].Fields.Contains(e.Field))
                {
                    foreach (var field in dataGrid.FieldLayouts[index: 0].Fields)
                    {
                        if (field.Index < dataGrid.FieldLayouts[index: 1].Fields.Count)
                        {
                            var field2 = dataGrid.FieldLayouts[index: 1].Fields[field.Index];
                            field2.ActualPosition = new FieldPosition(field.ActualPosition.Column,
                                                                      field.ActualPosition.Row,
                                                                      field.ActualPosition.ColumnSpan,
                                                                      field.ActualPosition.RowSpan
                                                                     );
                        }
                    }
                }

                if (dataGrid.FieldLayouts[index: 1].Fields.Contains(e.Field))
                {
                    foreach (var field in dataGrid.FieldLayouts[index: 1].Fields)
                    {
                        var field2 = dataGrid.FieldLayouts[index: 0].Fields[field.Index];
                        field2.ActualPosition = new FieldPosition(field.ActualPosition.Column,
                                                                  field.ActualPosition.Row,
                                                                  field.ActualPosition.ColumnSpan,
                                                                  field.ActualPosition.RowSpan
                                                                 );
                    }
                }

                Dispatcher.InvokeAsync(() =>
                                       {
                                           // reset all the length on the children to match the indent from the parent
                                           for (int i = 0; i < dataGrid.FieldLayouts[index: 1].Fields.Count; ++i)
                                           {
                                               var widthAdjustment = (dataGrid.FieldLayouts[index: 0].Fields[i].ActualPosition.Column == 1) ? 20 : 0;

                                               var f = dataGrid.FieldLayouts[index: 1].Fields[i];
                                               f.Width = new FieldLength(dataGrid.FieldLayouts[index: 0].Fields[i].CellWidthResolved - widthAdjustment);
                                           }
                                       }
                                      );
            }

            SaveCurrentWorkspaceLayout();
        }

        private void OnLabelPresenterSizeChanged(object sender, SizeChangedEventArgs e)
        {
            if (sender is LabelPresenter lp)
            {
                if (lp.FindAncestorOfType(typeof(XamDataGrid)) is XamDataGrid { FieldLayouts: { Count: > 1 } } dataGrid) 
                {
                    if ((dataGrid.FieldLayouts[index: 0].Fields.Count >= lp.Field.Index) &&
                        (dataGrid.FieldLayouts[index: 1].Fields.Count >= lp.Field.Index))
                    {
                        var widthAdjustment = (lp.Field.Index == 1) ? 20 : 0;

                        if (dataGrid.FieldLayouts[index: 0].Fields.Contains(lp.Field))
                        {
                            var f = dataGrid.FieldLayouts[index: 1].Fields[lp.Field.Index];
                            f.Width = new FieldLength(lp.Field.CellWidthResolved - widthAdjustment);
                        }

                        if (dataGrid.FieldLayouts[index: 1].Fields.Contains(lp.Field))
                        {
                            var f = dataGrid.FieldLayouts[index: 0].Fields[lp.Field.Index];
                            f.Width = new FieldLength(lp.Field.CellWidthResolved + widthAdjustment);
                        }
                    }
                }   
            }

            SaveCurrentWorkspaceLayout();
        }

        private void DataGrid_OnPreviewMouseMove(object sender, MouseEventArgs e)
        {
            var lp = Utilities.GetAncestorFromType(e.OriginalSource as DependencyObject,
                                                   typeof(LabelPresenter),
                                                   allowSubclassOfType: true
                                                  ) as LabelPresenter;

            if (lp == null)
            {
                if (Utilities.GetAncestorFromType(e.OriginalSource as DependencyObject,
                                                  typeof(CellValuePresenter),
                                                  allowSubclassOfType: true
                                                 ) is CellValuePresenter cvp)
                    cvp.Field.Settings.AllowResize = false;
            }
            else
                lp.Field.Settings.AllowResize = true;

            SaveCurrentWorkspaceLayout();
        }

        private void SaveCurrentWorkspaceLayout()
        {
            Dispatcher.InvokeAsync(() =>
                                   {
                                       var dataGrid = this.FindChildType<XamDataGrid>();

                                       if (dataGrid != null)
                                           CurrentWorkspace?.SaveLayout(dataGrid);
                                   }
                                  );
        }

        private void DragSource_OnDrop(object sender, DropEventArgs e)
        {
            Workspace sourceWorkspace = CurrentWorkspace;

            if ((Utilities.GetAncestorFromType(e.DropTarget, typeof(TabItemEx), allowSubclassOfType: false) as TabItemEx)?.DataContext is Workspace workspace && sourceWorkspace != null && sourceWorkspace != workspace)
            {
                if ((Utilities.GetAncestorFromType(e.DragSource, typeof(DataRecordPresenter), allowSubclassOfType: false) as DataRecordPresenter)?.Record is DataRecord { DataItem: DataRow row })
                {
                    // move it to the dragged tab
                    row.AddWorkspace(workspace.WorkspaceId);
                    row.Children?.ForEach(x => x.AddWorkspace(workspace.WorkspaceId));

                    if (Keyboard.IsKeyDown(Key.LeftCtrl) || Keyboard.IsKeyDown(Key.RightCtrl))
                    {
                        // do nothing its a copy
                    }
                    else
                    {
                        row.RemoveWorkspace(sourceWorkspace.WorkspaceId);
                        row.Children?.ForEach(x => x.RemoveWorkspace(sourceWorkspace.WorkspaceId));
                    }

                    SetData(CurrentWorkspace);
                }
            }
        }

        private void OnMouseButtonBlock(object sender, MouseButtonEventArgs e)
        {
            e.Handled = true;
        }

        private void ToggleButtonChanged(object sender, RoutedEventArgs e)
        {
            if (sender is DependencyObject so)
            {
                if (so.FindAncestorOfType(typeof(CellValuePresenter)) is CellValuePresenter cvp)
                {
                    var cellRecord = cvp.Record;

                    if (cellRecord != null)
                    {
                        var property = cellRecord.DataItem is DataRow di ? di.GetType().GetProperty(cvp.Field.Name) : null;

                        var grid = cvp.FindParent<XamDataGrid>();

                        if (grid != null)
                        {
                            var records = grid.SelectedItems.ToArray();

                            if (records.Length > 0 && property != null)
                            {
                                var value = property.GetValue(cellRecord.DataItem);

                                foreach (var r in records)
                                {
                                    var dr = r as DataRecord;

                                    if (ReferenceEquals(r, cellRecord) == false)
                                    {
                                        Dispatcher?.InvokeAsync(() =>
                                                                {
                                                                    if (dr?.DataItem is DataRow recordItem)
                                                                        property.SetValue(recordItem, value);
                                                                },
                                                                DispatcherPriority.Background
                                                               );
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }

        private void OnDataGridSorted(object sender, SortedEventArgs e)
        {
            if (sender is XamDataGrid dataGrid)
            {
                var sort = new List<FieldSortDescription>();

                try
                {
                    var sourceLayout = dataGrid.FieldLayouts[index: 0];
                    var destLayout = dataGrid.FieldLayouts[index: 1];

                    // reset all the length on the children to match the indent from the parent
                    for (int i = 0; i < destLayout.Fields.Count; ++i)
                    {
                        var source = sourceLayout.Fields[i];

                        var idx = sourceLayout.SortedFields.IndexOf(source, isGroupBy: false);
                        if (idx >= 0)
                        {
                            // got a sorting field get the descriptor...
                            var desc = sourceLayout.SortedFields[idx];

                            // then build the same for the children...
                            var dest = new FieldSortDescription { Field = destLayout.Fields[i], Direction = desc.Direction, IsGroupBy = desc.IsGroupBy };

                            sort.Add(dest);
                        }
                    }

                    // clear the old sort
                    destLayout.SortedFields.Clear();

                    // make the new sort
                    sort.ForEach(x => destLayout.SortedFields.Add(x));
                }
                catch
                {
                    //TODO: Log exception if needed
                }
            }
        }
    }
}
