﻿using System.Collections.Generic;
using System.Windows;
using System.Windows.Media;

namespace WpfApp1
{
    public static class VisualTreeExtensions
    {
        public static T FindFirst<T>(FrameworkElement element) where T : FrameworkElement
        {
            T res = null;

            if (element != null)
            {
                res = element as T;

                if (res == null)
                {
                    int childrenCount = VisualTreeHelper.GetChildrenCount(element);

                    for (int i = 0; i < childrenCount; i++)
                    {
                        var child = VisualTreeHelper.GetChild(element, i) as FrameworkElement;

                        res = FindFirst<T>(child);

                        if (res != null)
                            break;
                    }
                }
            }

            return res;
        }

        public static T FindParent<T>(this DependencyObject child) where T : DependencyObject
        {
            //get parent item
            DependencyObject parentObject = VisualTreeHelper.GetParent(child);

            //we've reached the end of the tree
            if (parentObject == null) return null;

            //check if the parent matches the type we're looking for
            if (parentObject is T parent)
                return parent;
            else
                return FindParent<T>(parentObject);
        }

        public static T FindChildType<T>(this DependencyObject depObj)
            where T : DependencyObject
        {
            T result = null;


            if (depObj != null)
            {
                for (int i = 0; i < VisualTreeHelper.GetChildrenCount(depObj); i++)
                {
                    var child = VisualTreeHelper.GetChild(depObj, i);

                    result = child as T ?? FindChildType<T>(child);
                    if (result != null)
                        break;
                }
            }

            return result;
        }

        public static T FindChild<T>(DependencyObject parent, string childName)
            where T : DependencyObject
        {
            // Confirm parent and childName are valid. 
            if (parent == null) return null;

            T foundChild = null;

            int childrenCount = VisualTreeHelper.GetChildrenCount(parent);
            for (int i = 0; foundChild == null && i < childrenCount; i++)
            {
                var child = VisualTreeHelper.GetChild(parent, i);
                // If the child is not of the request child type child
                if (child is not T)
                {
                    // recursively drill down the tree
                    foundChild = FindChild<T>(child, childName);

                    // If the child is found, break so we do not overwrite the found child. 
                    if (foundChild != null) break;
                }
                else
                {
                    if (!string.IsNullOrEmpty(childName))
                    {
                        // If the child's name is set for search
                        if (child is FrameworkElement frameworkElement && frameworkElement.Name == childName)
                        {
                            // if the child's name is of the request name
                            foundChild = (T) child;
                        }
                        else
                            foundChild = FindChild<T>(child, childName);
                    }
                    else
                    {
                        // child element found.
                        foundChild = (T) child;
                    }
                }
            }

            return foundChild;
        }

        public static IEnumerable<T> FindVisualChildren<T>(this DependencyObject depObj)
            where T : DependencyObject
        {
            if (depObj != null)
            {
                for (int i = 0; i < VisualTreeHelper.GetChildrenCount(depObj); i++)
                {
                    DependencyObject child = VisualTreeHelper.GetChild(depObj, i);
                    if (child is T variable)
                        yield return variable;

                    foreach (T childOfChild in FindVisualChildren<T>(child))
                        yield return childOfChild;
                }
            }
        }

    }
}
