﻿using Infragistics.Windows.DataPresenter;
using JetBrains.Annotations;
using System;
using System.ComponentModel;
using System.IO;
using System.Runtime.CompilerServices;
using System.Text;
using System.Xml;
using System.Xml.Schema;
using System.Xml.Serialization;

namespace WpfApp1
{
    public class Workspace : INotifyPropertyChanged, IXmlSerializable
    {
        public Workspace(int id)
        {
            if (id <= 0)
                throw new ArgumentOutOfRangeException(nameof(id));

            WorkspaceId = id;
            Name = $"Workspace {id}";
        }

        #region Name property

        private string mName;

        public string Name
        {
            get => mName;
            set
            {
                if (mName != value)
                {
                    mName = value;

                    InvokePropertyChanged();
                }
            }
        }

        #endregion

        #region WorkspaceId readonly property

        private int mWorkspaceId;

        public int WorkspaceId
        {
            get => mWorkspaceId;
            private set
            {
                if (mWorkspaceId != value)
                {
                    mWorkspaceId = value;

                    InvokePropertyChanged();
                }
            }
        }

        #endregion

        #region LayoutCache readonly property

        private string mLayoutCache;

        public string LayoutCache
        {
            get => mLayoutCache;
            private set
            {
                if (mLayoutCache != value)
                {
                    mLayoutCache = value;

                    InvokePropertyChanged();
                }
            }
        }

        #endregion

        #region AllowRowFiltering property

        private bool mAllowRowFiltering = true;

        public bool AllowRowFiltering
        {
            get => mAllowRowFiltering;
            set
            {
                if (mAllowRowFiltering != value)
                {
                    mAllowRowFiltering = value;

                    InvokePropertyChanged();
                }
            }
        }

        #endregion

        public void SaveLayout(XamDataGrid grid)
        {
            if (grid == null)
                throw new ArgumentNullException(nameof(grid));

            using var ms = new MemoryStream();

            grid.SaveCustomizations(ms, saveFieldPositionInfoAlways: true);

            LayoutCache = Encoding.UTF8.GetString(ms.ToArray());
        }
		
        public event PropertyChangedEventHandler PropertyChanged;

        [NotifyPropertyChangedInvocator]
        private void InvokePropertyChanged([CallerMemberName] string propertyName = null)
        {
            PropertyChanged?.Invoke(this, new PropertyChangedEventArgs(propertyName));
        }

        #region Implementation of IXmlSerializable

        /// <summary>This method is reserved and should not be used. When implementing the <see langword="IXmlSerializable" /> interface, you should return <see langword="null" /> (<see langword="Nothing" /> in Visual Basic) from this method, and instead, if specifying a custom schema is required, apply the <see cref="T:System.Xml.Serialization.XmlSchemaProviderAttribute" /> to the class.</summary>
        /// <returns>An <see cref="T:System.Xml.Schema.XmlSchema" /> that describes the XML representation of the object that is produced by the <see cref="M:System.Xml.Serialization.IXmlSerializable.WriteXml(System.Xml.XmlWriter)" /> method and consumed by the <see cref="M:System.Xml.Serialization.IXmlSerializable.ReadXml(System.Xml.XmlReader)" /> method.</returns>
        public XmlSchema GetSchema() => null;

        /// <summary>Generates an object from its XML representation.</summary>
        /// <param name="reader">The <see cref="T:System.Xml.XmlReader" /> stream from which the object is deserialized.</param>
        public void ReadXml(XmlReader reader)
        {
            if (reader == null)
                throw new ArgumentNullException(nameof(reader));

            reader.MoveToContent();

            bool isEmpty = reader.IsEmptyElement;

            reader.ReadStartElement(nameof(Workspace));
            if (isEmpty == false)
            {
                AllowRowFiltering = bool.TryParse(reader.ReadElementString(nameof(AllowRowFiltering)), out var res) && res;
                LayoutCache = reader.ReadElementString(nameof(LayoutCache));

                reader.ReadEndElement();
            }
        }

        /// <summary>Converts an object into its XML representation.</summary>
        /// <param name="writer">The <see cref="T:System.Xml.XmlWriter" /> stream to which the object is serialized.</param>
        public void WriteXml(XmlWriter writer)
        {
            if (writer == null)
                throw new ArgumentNullException(nameof(writer));

            writer.WriteStartElement(nameof(Workspace));
            writer.WriteElementString(nameof(AllowRowFiltering), AllowRowFiltering.ToString());
            writer.WriteElementString(nameof(LayoutCache), LayoutCache);
            writer.WriteEndElement();
        }

        #endregion
    }
}
